import json
from typing import Union, List, Optional

import aiomcrcon
from nonebot import logger
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent
from nonebot_plugin_guild_patch import GuildMessageEvent
from .data_model import RconClient, RconCLIENTS
from ..common import get_member_nickname, send_msg_to_qq_common


async def get_rcon_clients(
        event: Union[GroupMessageEvent, GuildMessageEvent],
        server_list: list
) -> List[RconClient]:
    """获取 服务器名、ws客户端、Rcon连接"""
    res: List[RconClient] = []
    for per_client in RconCLIENTS:
        for per_server in server_list:
            # 如果 服务器名 == ws客户端中记录的服务器名，且ws客户端存在
            if per_server.server_name == per_client.server_name and per_client.rcon:
                if isinstance(event, GroupMessageEvent):
                    if event.group_id in per_server['group_list']:
                        res.append(per_client)
                if isinstance(event, GuildMessageEvent):
                    for per_guild in per_server.guild_list:
                        if per_guild.guild_id == event.guild_id and per_guild.channel_id == event.channel_id:
                            res.append(per_client)
    return res


async def remove_rcon_client(server_name: str):
    """移除 Rcon 客户端"""
    for client in RconCLIENTS:
        if client.server_name == server_name:
            await client.rcon.close()
            RconCLIENTS.remove(client)
            logger.error(f"[MC_QQ_Rcon]丨[Server:{server_name}] 的 WebSocket、Rcon 连接已断开")


async def rcon_send_msg_to_qq(
        bot: Bot,
        message: str,
        server_list: list,
        display_server_name: Optional[bool] = False

):
    await send_msg_to_qq_common(
        bot=bot,
        message=message,
        server_list=server_list,
        display_server_name=display_server_name,
        plugin_name="MC_QQ_Rcon"
    )


async def rcon_connect(client: aiomcrcon.Client, server_name: str):
    """连接 Rcon"""
    try:
        await client.connect()
    except aiomcrcon.RCONConnectionError as e:
        logger.error(f"[MC_QQ_Rcon]丨[Server:{server_name}] 的Rcon连接失败：{str(e)}")
    except aiomcrcon.IncorrectPasswordError as e:
        logger.error(f"[MC_QQ_Rcon]丨[Server:{server_name}] 的Rcon密码错误：{str(e)}")
    else:
        logger.success(f"[MC_QQ_Rcon]丨[Server:{server_name}] 的Rcon连接成功")


async def msg_process_to_cmd(
        bot: Bot,
        event: Union[GroupMessageEvent, GuildMessageEvent],
        send_group_name: Optional[bool] = False
):
    """消息处理为 命令"""
    # 获取昵称
    member_nickname = await get_member_nickname(bot, event, event.user_id)

    # 初始化日志消息
    text_msg = member_nickname + " 说："

    command_msg = "tellraw @p "

    message_list = [
        {"text": "[MC_QQ] ", "color": "yellow"},
    ]
    if send_group_name:
        if isinstance(event, GroupMessageEvent):
            message_list.append(
                {"text": (await bot.get_group_info(group_id=event.group_id))['group_name'] + " ", "color": "aqua"})
        elif isinstance(event, GuildMessageEvent):
            guild_name = (await bot.get_guild_meta_by_guest(guild_id=event.guild_id))['guild_name']
            for per_channel in (await bot.get_guild_channel_list(guild_id=event.guild_id, no_cache=True)):
                if str(event.channel_id) == per_channel['channel_id']:
                    message_list.append({"text": guild_name + "丨" + per_channel['channel_name'] + " ", "color": "aqua"})
                    break
    message_list.append({"text": member_nickname, "color": "aqua"})
    message_list.append({"text": " 说：", "color": "yellow"})

    for msg in event.message:
        # 文本
        if msg.type == "text":
            msg_dict = {"text": msg.data['text'].replace("\r", "").replace("\n", "\n * ") + " ", "color": "white"}
            text_msg += msg.data['text'].replace("\r", "").replace("\n", "\n * ")
        # 图片
        elif msg.type == "image":
            msg_dict = {"text": "[图片] ", "color": "yellow",
                        "clickEvent": {"action": "open_url", "value": msg.data['url']},
                        "hoverEvent": {"action": "show_text", "contents": [{"text": "查看图片", "color": "gold"}]}
                        }
            text_msg += '[图片]'
        # 表情
        elif msg.type == "face":
            msg_dict = {"text": "[表情] ", "color": "gold"}
            text_msg += '[表情]'
        # 语音
        elif msg.type == "record":
            msg_dict = {"text": "[语音] ", "color": "light_purple"}
            text_msg += '[语音]'
        # 视频
        elif msg.type == "video":
            msg_dict = {"text": "[视频] ", "color": "light_purple",
                        "clickEvent": {"action": "open_url", "value": msg.data['url']},
                        "hoverEvent": {"action": "show_text", "contents": [{"text": "查看视频", "color": "dark_purple"}]}
                        }
            text_msg += '[视频]'
        # @
        elif msg.type == "at":
            # 获取被@ 群/频道 昵称
            at_member_nickname = await get_member_nickname(bot, event, msg.data['qq'])
            msg_dict = {"text": "@" + at_member_nickname + " ", "color": "green"}
            text_msg += f"@{at_member_nickname}"
        # share
        elif msg.type == "share":
            msg_dict = {"text": "[分享：" + msg.data['title'] + "] ", "color": "yellow",
                        "clickEvent": {"action": "open_url", "value": msg.data['url']},
                        "hoverEvent": {"action": "show_text", "contents": [{"text": "查看图片", "color": "gold"}]}
                        }
            text_msg += '[分享：' + msg.data['title'] + ']'
        # forward
        elif msg.type == "forward":
            # TODO 将合并转发消息拼接为字符串
            # 获取合并转发 await bot.get_forward_msg(message_id=event.message_id)
            msg_dict = {"text": "[合并转发] ", "color": "white"}
            text_msg += '[合并转发]'
        else:
            msg_dict = {"text": "[ " + msg.type + "] ", "color": "white"}
            text_msg += '[' + msg.type + ']'

        # 放入消息列表
        message_list.append(msg_dict)

    # 拼接完整命令
    command_msg += json.dumps(message_list)
    return text_msg, command_msg


async def rcon_send_msg_to_mc(
        bot: Bot,
        event: Union[GroupMessageEvent, GuildMessageEvent],
        server_list: list
):
    """发送消息到 Minecraft"""
    text_msg, command_msg = await msg_process_to_cmd(bot=bot, event=event)
    if client_list := await get_rcon_clients(event=event, server_list=server_list):
        for client in client_list:
            if client and client.rcon:
                try:
                    await client.rcon.send_cmd(command_msg)
                except aiomcrcon.errors.ClientNotConnectedError as e:
                    logger.error(f"[MC_QQ_Rcon]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：{str(e)}")
                    # 连接关闭则移除客户端
                    await remove_rcon_client(server_name=client.server_name)
                else:
                    logger.success(f"[MC_QQ_Rcon]丨发送至 [server:{client.server_name}] 的消息 \"{text_msg}\"")


async def rcon_send_command_to_mc(
        bot: Bot,
        event: Union[GroupMessageEvent, GuildMessageEvent],
        server_list: list
):
    """发送命令到 Minecraft"""
    if client_list := await get_rcon_clients(event=event, server_list=server_list):
        for client in client_list:
            if client and client.rcon:
                try:
                    await bot.send(event, message=str(
                        (await client.rcon.send_cmd(
                            event.raw_message.strip("/").strip("mcc").strip())
                         )[0]))
                except aiomcrcon.errors.ClientNotConnectedError as e:
                    logger.error(f"[MC_QQ_Rcon]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：{str(e)}")
                    # 连接关闭则移除客户端
                    await remove_rcon_client(server_name=client.server_name)
                else:
                    logger.success(
                        f"[MC_QQ_Rcon]丨发送至 [server:{client.server_name}] 的命令 \"{event.raw_message.strip('/').strip('/mcc')}\""
                    )



from typing import Optional, List, Dict

from pydantic import BaseModel, Extra, Field


class Guild(BaseModel):
    """频道配置"""
    # 频道ID
    guild_id: int
    # 子频道ID
    channel_id: int


class Server(BaseModel):
    """服务器配置"""
    # 服务器名称
    server_name: str
    # 服务器群列表
    group_list: Optional[List[int]] = []
    # 服务器频道列表
    guild_list: Optional[List[Guild]] = []


class Config(BaseModel, extra=Extra.ignore):
    # 服务器地址
    mc_qq_ws_ip: Optional[str] = "127.0.0.1"
    # 服务器端口
    mc_qq_rcon_ws_port: Optional[int] = 8766
    # 是否发送群聊名称
    mc_qq_send_group_name: Optional[bool] = False
    # 是否显示服务器名称
    mc_qq_display_server_name: Optional[bool] = False
    # 服务器列表
    mc_qq_rcon_servers_list: Optional[List[Server]] = Field(default_factory=list)
    # MCRcon 密码
    mc_qq_rcon_password: Optional[str] = "password"
    # Rcon 字典
    mc_qq_rcon_rcon_dict: Optional[Dict[str, int]] = Field(default_factory=dict)



from typing import Optional
from pydantic import BaseModel
from .basemodel import BasePlayer, BaseChatEvent, BaseDeathEvent, BaseJoinEvent, BaseQuitEvent


class Address(BaseModel):
    address: str
    port: int


class SpigotPlayer(BasePlayer):
    """玩家信息"""
    uuid: str
    display_name: str
    player_list_name: str
    is_health_scaled: bool
    address: Address
    is_sprinting: bool
    walk_speed: float
    fly_speed: float
    is_sneaking: bool
    level: int
    is_flying: bool
    ping: Optional[int] = None
    """Spigot API 1.12.2 Player 无 ping 属性"""
    allow_flight: bool
    locale: str
    health_scale: float
    player_time_offset: int
    exp: float
    total_exp: int
    player_time: int
    is_player_time_relative: bool


class AsyncPlayerChatEvent(BaseChatEvent):
    """Spigot API AsyncPlayerChatEvent"""
    event_name = "AsyncPlayerChatEvent"
    player: SpigotPlayer


class PlayerJoinEvent(BaseJoinEvent):
    """Spigot API PlayerJoinEvent"""
    event_name = "PlayerJoinEvent"
    player: SpigotPlayer


class PlayerQuitEvent(BaseQuitEvent):
    """Spigot API PlayerQuitEvent"""
    event_name = "PlayerQuitEvent"
    player: SpigotPlayer


class PlayerDeathEvent(BaseDeathEvent):
    """Spigot API PlayerDeathEvent"""
    event_name = "PlayerDeathEvent"
    player: SpigotPlayer



from .minecraft import MineCraftPlayerChatEvent, MinecraftPlayerJoinEvent, MinecraftPlayerQuitEvent
from .spigot import AsyncPlayerChatEvent, PlayerDeathEvent, PlayerJoinEvent, PlayerQuitEvent

event_dict = {
    # 原版
    "MinecraftPlayerJoinEvent": MinecraftPlayerJoinEvent,
    "MinecraftPlayerQuitEvent": MinecraftPlayerQuitEvent,
    "MineCraftPlayerChatEvent": MineCraftPlayerChatEvent,
    # Spigot
    "AsyncPlayerChatEvent": AsyncPlayerChatEvent,
    "PlayerDeathEvent": PlayerDeathEvent,
    "PlayerJoinEvent": PlayerJoinEvent,
    "PlayerQuitEvent": PlayerQuitEvent,
}



import aiomcrcon
import websockets
from typing import List


class RconClient:
    """MC_QQ Rcon 客户端"""
    server_name: str
    websocket: websockets.WebSocketServerProtocol
    rcon: aiomcrcon.Client

    def __init__(self, server_name: str, websocket: websockets.WebSocketServerProtocol, rcon: aiomcrcon.Client):
        self.server_name: str = server_name
        self.websocket: websockets.WebSocketServerProtocol = websocket
        self.rcon: aiomcrcon.Client = rcon


RconCLIENTS: List[RconClient] = []
"""Rcon 客户端列表"""



from pydantic import BaseModel


class BaseEvent(BaseModel):
    """事件基类"""
    post_type: str
    event_name: str
    server_name: str


class BasePlayer(BaseModel):
    """玩家信息基类"""
    nickname: str


class BaseMessageEvent(BaseEvent):
    """消息事件基类"""
    post_type = "message"
    player: BasePlayer


class BaseChatEvent(BaseEvent):
    """玩家聊天事件基类"""
    post_type = "message"
    sub_type = "chat"
    message: str
    player: BasePlayer


class BaseDeathEvent(BaseEvent):
    """玩家死亡事件基类"""
    post_type = "message"
    sub_type = "death"
    player: BasePlayer
    death_message: str


class BaseNoticeEvent(BaseEvent):
    """通知事件基类"""
    post_type = "notice"


class BaseJoinEvent(BaseNoticeEvent):
    """玩家加入事件基类"""
    sub_type = "join"
    player: BasePlayer


class BaseQuitEvent(BaseNoticeEvent):
    """玩家退出事件基类"""
    sub_type = "quit"
    player: BasePlayer



from typing import List

import websockets


class Client:
    """MC_QQ Rcon 客户端"""
    server_name: str
    websocket: websockets.WebSocketServerProtocol

    def __init__(self, server_name: str, websocket: websockets.WebSocketServerProtocol):
        self.server_name: str = server_name
        self.websocket: websockets.WebSocketServerProtocol = websocket


CLIENTS: List[Client] = []



from typing import List, Union, Optional

import websockets
from nonebot import logger
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent
from nonebot_plugin_guild_patch import GuildMessageEvent
from .data_model import Client, CLIENTS
from ..common import get_member_nickname, send_msg_to_qq_common


async def send_msg_to_mc(
        bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent],
        server_list: list
):
    """发送消息到 MC"""
    # 处理来自QQ的消息
    text_msg, msgJson = await msg_process_to_json(bot=bot, event=event)
    if client_list := await get_clients(event=event, server_list=server_list):
        for client in client_list:
            if client and client.websocket:
                try:
                    await client.websocket.send(msgJson)
                except websockets.WebSocketException as e:
                    logger.error(f"[MC_QQ]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：{e}")
                    CLIENTS.remove(client)
                else:
                    logger.success(f"[MC_QQ]丨发送至 [server:{client.server_name}] 的消息 \"{text_msg}\"")


async def get_clients(
        event: Union[GroupMessageEvent, GuildMessageEvent],
        server_list: list
) -> List[Client]:
    """获取 服务器名、ws客户端, 返回client列表"""
    res: List[Client] = []
    for per_client in CLIENTS:
        for per_server in server_list:
            # 如果 服务器名 == ws客户端中记录的服务器名，且ws客户端存在
            if per_server.server_name == per_client.server_name and per_client.websocket and (
                    per_client not in res):
                if isinstance(event, GroupMessageEvent):
                    if event.group_id in per_server.group_list:
                        res.append(per_client)
                if isinstance(event, GuildMessageEvent):
                    for per_guild in per_server.guild_list:
                        if per_guild.guild_id == event.guild_id and per_guild.channel_id == event.channel_id:
                            res.append(per_client)
    return res


async def remove_client(server_name: str):
    """移除客户端"""
    for client in CLIENTS:
        if client.server_name == server_name:
            await client.websocket.close()
            CLIENTS.remove(client)


async def send_msg_to_qq(
        bot: Bot,
        message: str,
        server_list: list,
        display_server_name: Optional[bool] = False,
):
    """
    发送消息到 QQ
    :param bot: Bot
    :param message: 消息
    :param server_list: 服务器列表
    :param display_server_name: 是否显示服务器名
    """
    await send_msg_to_qq_common(
        bot=bot,
        message=message,
        server_list=server_list,
        display_server_name=display_server_name,
        plugin_name="MC_QQ"
    )


async def msg_process_to_json(
        bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent],
        send_group_name: Optional[bool] = False
):
    """消息处理为 JSON"""
    # 获取昵称
    member_nickname = await get_member_nickname(bot, event, event.user_id)

    # 初始化消息
    text_msg = member_nickname + "说："

    # 初始化消息字典
    messageList = []

    # 发送群聊名称
    if send_group_name:
        group_name = {'msgType': "group_name"}
        if isinstance(event, GroupMessageEvent):
            group_name['msgData'] = (await bot.get_group_info(group_id=event.group_id))['group_name']
        elif isinstance(event, GuildMessageEvent):
            guild_name = (await bot.get_guild_meta_by_guest(guild_id=event.guild_id))['guild_name']
            for per_channel in (await bot.get_guild_channel_list(guild_id=event.guild_id, no_cache=True)):
                if str(event.channel_id) == per_channel['channel_id']:
                    channel_name = per_channel['channel_name']
                    group_name['msgData'] = f"{guild_name}丨{channel_name}"
                    break
        messageList.append({"msgType": "group_name", "msgData": group_name})

    # 将群成员昵称装入消息列表
    messageList.append({"msgType": "senderName", "msgData": member_nickname})

    for msg in event.message:
        per_msg = {'msgType': msg.type}
        # 文本
        if msg.type == "text":
            msgData = msg.data['text'].replace("\r", "").replace("\n", "\n * ")
            text_msg += msgData
        # 图片
        elif msg.type == "image":
            msgData = msg.data['url']
            text_msg += '[图片]'
        # 表情
        elif msg.type == "face":
            msgData = '[表情]'
            text_msg += '[表情]'
        # 语音
        elif msg.type == "record":
            msgData = '[语音]'
            text_msg += '[语音]'
        # 视频
        elif msg.type == "video":
            msgData = msg.data['url']
            text_msg += '[视频]'
        # @
        elif msg.type == "at":
            # 获取被@ 群/频道 昵称
            at_member_nickname = await get_member_nickname(bot, event, msg.data['qq'])
            msgData = f"@{at_member_nickname}"
            text_msg += msgData
        # share
        elif msg.type == "share":
            msgData = msg.data['url']
            text_msg += '[分享：' + msg.data['title'] + ']'
        # forward
        elif msg.type == "forward":
            # TODO 将合并转发消息拼接为字符串
            # 获取合并转发 await bot.get_forward_msg(message_id=event.message_id)
            msgData = '[合并转发]'
            text_msg = msgData
        else:
            msgData = msg.type
            text_msg += '[' + msg.type + '] '

        text_msg += " "

        # 装入消息数据
        per_msg['msgData'] = msgData
        # 放入消息列表
        messageList.append(per_msg)

    return text_msg, str({"message": messageList})



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="mcqq-tool",  # 项目名称，保证它的唯一性，不要跟已存在的包名冲突即可
    version="0.0.1",  # 程序版本
    author="17TheWord",  # 项目作者
    author_email="17theword@gmail.com",  # 作者邮件
    description="MC_QQ 工具包",  # 项目的一句话描述
    long_description=long_description,  # 加长版描述？
    long_description_content_type="text/markdown",  # 描述使用Markdown
    url="https://github.com/17TheWord/mcqq-tool",  # 项目地址
    packages=setuptools.find_packages(),  # 无需修改
    classifiers=[
        "Programming Language :: Python :: 3.9",  # 使用Python3.9
        "License :: OSI Approved :: GNU Affero General Public License v3",  # 开源协议
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'websockets>=10.3',
        'aio-mc-rcon>=3.2.0'
    ]
)



from .config import Config, Server
from .data_model import RconClient, RconCLIENTS
from .utils import rcon_send_msg_to_qq, rcon_send_msg_to_mc, remove_rcon_client, rcon_send_command_to_mc, rcon_connect
